"""
=============================
Method Validation for Metrics
=============================
Various methods and objects for validating metric inputs.

Functions
---------
 - raise_for_non_vector
 - raise_for_inconsistent_shapes
 - warn_for_nondichotomous_categories

Classes
-------
 - InconsistentShapesError
 - NonVectorError

"""

import numpy as np
import numpy.typing as npt
from typing import List, Tuple, Sequence
import pandas as pd
import warnings
from pandas.api.types import CategoricalDtype

class InconsistentShapesError(Exception):
    def __init__(self,
        *args: object,
        array_shape_1: Tuple[int],
        array_shape_2: Tuple[int]
        ) -> None:
        self._array_shape_1 = array_shape_1
        self._array_shape_2 = array_shape_2
        super().__init__(*args)

    def __str__(self) -> str:
        a1s = self._array_shape_1
        a2s = self._array_shape_2
        message = f"Arrays are not the same shape. The array shapes are {a1s} and {a2s}."
        return message

class NonVectorError(Exception):
    def __init__(self,
        *args: object,
        arr: npt.ArrayLike
        ) -> None:
        self._arr = arr
        super().__init__(*args)

    def __str__(self) -> str:
        message = f"This array is not 1-dimensional\n {self._arr}"
        return message


def _array_attr(seq: Sequence, attr: str) -> npt.ArrayLike:
    if not hasattr(seq, attr):
        return np.asarray(seq)
    return seq

def raise_for_non_vector(
    *arrays: List[npt.ArrayLike]
    ) -> None:
    """
    Validate that all arrays are 1-dimensional.

    Parameters
    ----------
    arrays: arbitrary number of array-like values, required
        Arrays to check.

    Raises
    ------
    NonVectorError:
        Raises if one of the arrays is not 1-dimensional.
    """
    # Check each array
    for x in arrays:
        if _array_attr(x, "ndim").ndim != 1:
            raise NonVectorError(arr=x)

def raise_for_inconsistent_shapes(
    *arrays: List[npt.ArrayLike]
    ) -> None:
    """
    Validate that all arrays are the same shape.

    Parameters
    ----------
    arrays: arbitrary number of array-like values, required
        Arrays to check.

    Raises
    ------
    InconsistentShapesError:
        Raises if all of the arrays are not the same shape.
    """
    # Extract first array
    xshape = _array_attr(arrays[0], "shape").shape

    # Check shape of each
    for y in arrays[1:]:
        yshape = _array_attr(y, "shape").shape
        if xshape != yshape:
            raise InconsistentShapesError(
                array_shape_1=xshape,
                array_shape_2=yshape
                )

def convert_to_boolean_categorical_series(
    data: npt.ArrayLike
    ) -> pd.Series:
    """
    Transform data into a boolean categorical pandas.Series.

    Parameters
    ----------
    data: array-like, required
        Data to convert. Should only contain True or False values.

    Warnings
    --------
    UserWarning:
        Warns if any values in data are not True or False. These values will become NaN.

    Returns
    -------
    Validated boolean categorical series.
    """
    # Create boolean categorical series
    s = pd.Series(
        data=data,
        dtype=CategoricalDtype([True, False])
    )

    # Check for NaN
    if s.isnull().any():
        message = f"{data} contains values that could not be converted to True or False."
        warnings.warn(message=message, category=UserWarning)

    return s
    



import pytest
import numpy as np
from hydrotools.metrics import _validation
import pandas as pd

def test_raise_for_non_vector():
    x = np.array([[1, 2, 3, 4], [1, 1, 1, 4]])
    y = [1, 2, 3, 4]

    _validation.raise_for_non_vector(y)

    with pytest.raises(_validation.NonVectorError):
        _validation.raise_for_non_vector(x)

def test_raise_for_inconsistent_shapes():
    x = np.array([[1, 2, 3, 4], [1, 1, 1, 4]])
    y = [1, 2, 3, 4]
    z = [5, 6, 7, 8]

    _validation.raise_for_inconsistent_shapes(y, z)

    with pytest.raises(_validation.InconsistentShapesError):
        _validation.raise_for_inconsistent_shapes(x, y)

def test_convert_to_boolean_categorical_series():
    # Check for non-Series
    with pytest.warns(UserWarning):
        x = _validation.convert_to_boolean_categorical_series([1, 2, 3])

    # Check for non-categorical
    s = pd.Series([True, True, False])
    s = _validation.convert_to_boolean_categorical_series(s)
    assert hasattr(s, "cat")

    # Check for True
    s = pd.Series([False, False, False], dtype="category")
    s = _validation.convert_to_boolean_categorical_series(s)
    assert True in s.cat.categories

    # Check for False
    s = pd.Series([False, False, False], dtype="category")
    s = _validation.convert_to_boolean_categorical_series(s)
    assert False in s.cat.categories

    # Check for two categories
    with pytest.warns(UserWarning):
        s = pd.Series([True, False, "5"], dtype="category")
        s = _validation.convert_to_boolean_categorical_series(s)
        assert len(s.cat.categories) == 2
        assert True in s.cat.categories
        assert False in s.cat.categories



import pytest
from hydrotools.metrics import metrics

import pandas as pd

all_tp_case = pd.DataFrame({
    "obs": pd.Categorical([True, True, True]),
    "sim": pd.Categorical([True, True, True])
})
all_fp_case = pd.DataFrame({
    "obs": pd.Categorical([False, False, False]),
    "sim": pd.Categorical([True, True, True])
})
all_fn_case = pd.DataFrame({
    "obs": pd.Categorical([True, True, True]),
    "sim": pd.Categorical([False, False, False])
})
all_tn_case = pd.DataFrame({
    "obs": pd.Categorical([False, False, False]),
    "sim": pd.Categorical([False, False, False])
})

scenarios = [
    (all_tp_case, "true_positive", 3),
    (all_fp_case, "false_positive", 3),
    (all_fn_case, "false_negative", 3),
    (all_tn_case, "true_negative", 3)
]

@pytest.mark.parametrize("data,check,value", scenarios)
def test_compute_contingency_table_scenarios(data, check, value):
    # Construct contingency table
    table = metrics.compute_contingency_table(data["obs"], data["sim"])

    # Validate correct values
    for component, val in table.items():
        if component == check:
            assert val == value
        else:
            assert val == 0

def test_non_series():
    obs = [True, False, True, False]
    sim = [True, True, True, True]

    table = metrics.compute_contingency_table(obs, sim)
    assert table["true_positive"] == 2
    assert table["false_positive"] == 2
    assert table["false_negative"] == 0
    assert table["true_negative"] == 0

    POD = metrics.probability_of_detection(table)
    assert POD == 1.0

    POFD = metrics.probability_of_false_detection(table)
    assert POFD == 1.0



"""
==================
Evaluation Metrics
==================
Convenience methods for computing common evaluation metrics.

For a description of common evaluation metrics, see:

http://www.eumetrain.org/data/4/451/english/courses/msgcrs/index.htm

Functions
---------
 - compute_contingency_table
 - probability_of_detection
 - probability_of_false_detection
 - probability_of_false_alarm
 - threat_score
 - frequency_bias
 - percent_correct
 - base_chance
 - equitable_threat_score
 - mean_error
 - nash_sutcliffe_efficiency
 - kling_gupta_efficiency
 - volumetric_efficiency
 - mean_squared_error
 - root_mean_squared_error
 - mean_error_skill_score
 - coefficient_of_persistence
 - coefficient_of_extrapolation

"""

import numpy as np
import numpy.typing as npt
import pandas as pd
from typing import Union
from . import _validation as validate
from functools import partial

def mean_error(
    y_true: npt.ArrayLike,
    y_pred: npt.ArrayLike,
    power: float = 1.0,
    root: bool = False
    ) -> float:
    r"""Compute the mean error or deviation. Default is Mean Absolute Error. The mean error 
    is given by:

    $$ME = \frac{1}{n}\sum_{i=1}^{n}\left| y_{s,i} - y_{o,i} \right|^{p}$$

    Where $n$ is the length of each array, $y_{s,i}$ is the *ith* simulated or predicted value, 
    $y_{o,i}$ is the *ith* observed or true value, and $p$ is the exponent.
        
    Parameters
    ----------
    y_true: array-like of shape (n_samples,), required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: array-like of shape (n_samples,), required
        Estimated target values, also called simulations or modeled values.
    power: float, default 1.0
        Exponent for each mean error summation value.
    root: bool, default False
        When True, return the root mean error.
        
    Returns
    -------
    mean_error: float
        Mean error or root mean error.
    
    """
    # Compute mean error
    _mean_error = np.sum(np.abs(np.subtract(y_true, y_pred)) ** power) / len(y_true)

    # Return mean error, optionally return root mean error
    if root:
        return np.sqrt(_mean_error)
    return _mean_error

mean_squared_error = partial(mean_error, power=2.0, root=False)
mean_squared_error.__doc__ = """Partial of hydrotools.metrics.mean_error with 
a default power value of 2.0 and root set to False.

See Also
--------
mean_error
"""

root_mean_squared_error = partial(mean_error, power=2.0, root=True)
root_mean_squared_error.__doc__ = """Partial of hydrotools.metrics.mean_error with 
a default power value of 2.0 and root set to True.

See Also
--------
mean_error
"""

def mean_error_skill_score(
    y_true: npt.ArrayLike,
    y_pred: npt.ArrayLike,
    y_base: npt.ArrayLike,
    power: float = 1.0,
    normalized: bool = False
    ) -> float:
    r"""Compute a generic mean error based model skill score. The mean error skill score 
    is given by:

    $$MESS = 1 - \frac{\sum_{i=1}^{n}\left| y_{p,i} - y_{o,i} \right|^{p}}{\sum_{i=1}^{n}\left| y_{b,i} - y_{o,i} \right|^{p}}$$

    Where $n$ is the length of each array, $y_{s,i}$ is the *ith* simulated or predicted value, 
    $y_{b,i}$ is the *ith* baseline value, $y_{o,i}$ is the *ith* observed or true 
    value, and $p$ is the exponent.
        
    Parameters
    ----------
    y_true: array-like of shape (n_samples,), required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: array-like of shape (n_samples,), required
        Estimated target values, also called simulations or modeled values.
    y_base: array-like of shape (n_samples,), required
        Baseline value(s) against which to assess skill of y_pred.
    power: float, default 1.0
        Exponent for each mean error summation value.
    normalized: bool, default False
        When True, normalize the final skill score using the method from 
        Nossent & Bauwens, 2012.
        
    Returns
    -------
    score: float
        Skill score of y_pred relative to y_base.
        
    References
    ----------
    Nossent, J., & Bauwens, W. (2012, April). Application of a normalized 
        Nash-Sutcliffe efficiency to improve the accuracy of the Sobol'sensitivity 
        analysis of a hydrological model. In EGU General Assembly Conference 
        Abstracts (p. 237).
    
    """
    # Compute components
    numerator = mean_error(y_true, y_pred, power=power)
    denominator = mean_error(y_true, y_base, power=power)

    # Compute score, optionally normalize
    if normalized:
        return 1.0 / (1.0 + numerator/denominator)
    return 1.0 - numerator/denominator

volumetric_efficiency = partial(mean_error_skill_score, y_base=0.0, power=1.0)
volumetric_efficiency.__doc__ = """Partial of hydrotools.metrics.mean_error_skill_score 
with a default y_base of 0.0 and a power value of 1.0. Volumetric efficiency ranges from 
-inf to 1.0, higher is better. According to the authors, volumetric efficiency indicates 
the "portion of water that arrives on time." Note: that large over-predictions result in 
deeply negative values.

See Also
--------
mean_error_skill_score

References
----------
Criss, R. E., & Winston, W. E. (2008). Do Nash values have value? Discussion 
    and alternate proposals. Hydrological Processes: An International Journal, 
    22(14), 2723-2725.
"""

def nash_sutcliffe_efficiency(
    y_true: npt.ArrayLike,
    y_pred: npt.ArrayLike,
    log: bool = False,
    power: float = 2.0,
    normalized: bool = False
    ) -> float:
    """Compute the Nash-Sutcliffe model efficiency coefficient (NSE), also called the 
    mean squared error skill score or the R^2 (coefficient of determination) regression score. 
    The NSE compares model errors to observed variance. The default NSE ranges from -inf to 1.0, 
    higher is better. A score of 0.0 indicates the model is as good a predictor as the mean of 
    observations. A score of 1.0 indicates the model exactly matches the observations.

    The "normalized" Nash-Sutcliffe model efficiency re-scales the NSE to a range from 0.0 to 1.0. 
    In this case, A score of 0.5 indicates the model is as good a predictor as the mean of 
    observations. A score of 1.0 still indicates the model exactly matches the observations.
        
    Parameters
    ----------
    y_true: array-like of shape (n_samples,), required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: array-like of shape (n_samples,), required
        Estimated target values, also called simulations or modeled values.
    log: bool, default False
        Apply numpy.log (natural logarithm) to y_true and y_pred 
        before computing the NSE.
    power: float, default 2.0
        Exponent for each mean error summation value.
    normalized: bool, default False
        When True, normalize the final NSE value using the method from 
        Nossent & Bauwens, 2012.
        
    Returns
    -------
    score: float
        Nash-Sutcliffe model efficiency coefficient
        
    References
    ----------
    Nash, J. E., & Sutcliffe, J. V. (1970). River flow forecasting through 
        conceptual models part I—A discussion of principles. Journal of 
        hydrology, 10(3), 282-290.
    Nossent, J., & Bauwens, W. (2012, April). Application of a normalized 
        Nash-Sutcliffe efficiency to improve the accuracy of the Sobol'sensitivity 
        analysis of a hydrological model. In EGU General Assembly Conference 
        Abstracts (p. 237).
    
    """
    # Raise if not 1-D arrays
    validate.raise_for_non_vector(y_true, y_pred)

    # Raise if not same shape
    validate.raise_for_inconsistent_shapes(y_true, y_pred)

    # Optionally transform components
    if log:
        y_true = np.log(y_true)
        y_pred = np.log(y_pred)

    # Compute score
    return mean_error_skill_score(y_true, y_pred, np.mean(y_true), 
        power=power, normalized=normalized)

def coefficient_of_persistence(
    y_true: npt.ArrayLike,
    y_pred: npt.ArrayLike,
    lag: int = 1,
    log: bool = False,
    power: float = 2.0,
    normalized: bool = False
    ) -> float:
    """Compute the coefficient of persistence (Kitanidis & Bras, 1980). The coefficient of 
    persistence compares the model to a recent observation, given some lag. This score assesses 
    the model's skill compared to assuming a previous observation does not change (persists).

    In the default case, the *ith* modeled value will be compared to the *i-1* observed value. 
    The result is the mean squared error skill score using the *i-1* observed values as a baseline. 
    The coefficient of persistence ranges from -inf to 1.0, higher is better. A score of 0.0 indicates 
    skill no better than assuming the last observation would persist. A perfect score is 1.0.
        
    Parameters
    ----------
    y_true: array-like of shape (n_samples,), required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: array-like of shape (n_samples,), required
        Estimated target values, also called simulations or modeled values.
    lag: int, default 1
        Number of values by which to lag the baseline.
    log: bool, default False
        Apply numpy.log (natural logarithm) to y_true and y_pred 
        before computing the score.
    power: float, default 2.0
        Exponent for each mean error summation value.
    normalized: bool, default False
        When True, normalize the final score using the method from 
        Nossent & Bauwens, 2012.
        
    Returns
    -------
    score: float
        Coefficient of persistence.

    See Also
    --------
    mean_error_skill_score: Generic method for computing model skill.
        
    References
    ----------
    Kitanidis, P. K., & Bras, R. L. (1980). Real-time forecasting with a conceptual 
        hydrologic model: 2. Applications and results. Water Resources Research, 
        16(6), 1034-1044.
    Nossent, J., & Bauwens, W. (2012, April). Application of a normalized 
        Nash-Sutcliffe efficiency to improve the accuracy of the Sobol'sensitivity 
        analysis of a hydrological model. In EGU General Assembly Conference 
        Abstracts (p. 237).
    
    """
    # Raise if not 1-D arrays
    validate.raise_for_non_vector(y_true, y_pred)

    # Raise if not same shape
    validate.raise_for_inconsistent_shapes(y_true, y_pred)

    # Optionally transform components
    if log:
        y_true = np.log(y_true)
        y_pred = np.log(y_pred)

    # Compute baseline
    y_base = np.roll(y_true, lag)

    # Compute score
    return mean_error_skill_score(y_true[lag:], y_pred[lag:], y_base[lag:], 
        power=power, normalized=normalized)

def coefficient_of_extrapolation(
    y_true: npt.ArrayLike,
    y_pred: npt.ArrayLike,
    log: bool = False,
    power: float = 2.0,
    normalized: bool = False
    ) -> float:
    """Compute the coefficient of extrapolation (Kitanidis & Bras, 1980). The coefficient of 
    extrapolation compares the model output to the last two values of the observations, assuming 
    the linear trend of the these values will continue. In other words, the coefficient of 
    extrapolation is a skill score with baseline values $y_{b,i} = y_{b,i-1} + (y_{b,i-1} - y_{b,i-2})$.

    The coefficient of extrapolation ranges from -inf to 1.0, higher is better. A score of 0.0 indicates 
    skill no better than assuming the difference between the last two observations will persist. A perfect 
    score is 1.0.
        
    Parameters
    ----------
    y_true: array-like of shape (n_samples,), required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: array-like of shape (n_samples,), required
        Estimated target values, also called simulations or modeled values.
    log: bool, default False
        Apply numpy.log (natural logarithm) to y_true and y_pred 
        before computing the NSE.
    power: float, default 2.0
        Exponent for each mean error summation value.
    normalized: bool, default False
        When True, normalize the final NSE value using the method from 
        Nossent & Bauwens, 2012.
        
    Returns
    -------
    score: float
        Coefficient of extrapolation.

    See Also
    --------
    mean_error_skill_score: Generic method for computing model skill.
        
    References
    ----------
    Kitanidis, P. K., & Bras, R. L. (1980). Real-time forecasting with a conceptual 
        hydrologic model: 2. Applications and results. Water Resources Research, 
        16(6), 1034-1044.
    Nossent, J., & Bauwens, W. (2012, April). Application of a normalized 
        Nash-Sutcliffe efficiency to improve the accuracy of the Sobol'sensitivity 
        analysis of a hydrological model. In EGU General Assembly Conference 
        Abstracts (p. 237).
    
    """
    # Raise if not 1-D arrays
    validate.raise_for_non_vector(y_true, y_pred)

    # Raise if not same shape
    validate.raise_for_inconsistent_shapes(y_true, y_pred)

    # Optionally transform components
    if log:
        y_true = np.log(y_true)
        y_pred = np.log(y_pred)

    # Compute baseline
    slope = np.diff(y_true)[:-1]
    y_base = y_true[2:] + slope

    # Compute score
    return mean_error_skill_score(y_true[2:], y_pred[2:], y_base, 
        power=power, normalized=normalized)

def kling_gupta_efficiency(
    y_true: npt.ArrayLike,
    y_pred: npt.ArrayLike,
    r_scale: float = 1.0,
    a_scale: float = 1.0,
    b_scale: float = 1.0
    ) -> float:
    """Compute the Kling-Gupta model efficiency coefficient (KGE). The KGE is a 
    summary metric that combines the relative mean, relative variance, and linear 
    correlation between observed and simulated values. The final metric is computed 
    using the root sum of squares with optional scaling factors, similar to 
    computing distance in a 3-dimensional Euclidean space.
        
    Parameters
    ----------
    y_true: array-like of shape (n_samples,), required
        Ground truth (correct) target values, also called observations, measurements, or observed values.
    y_pred: array-like of shape (n_samples,), required
        Estimated target values, also called simulations or modeled values.
    r_scale: float, optional, default 1.0
        Linear correlation (r) scaling factor. Used to re-scale the Euclidean space by 
        emphasizing different KGE components.
    a_scale: float, optional, default 1.0
        Relative variability (alpha) scaling factor. Used to re-scale the Euclidean space by 
        emphasizing different KGE components.
    b_scale: float, optional, default 1.0
        Relative mean (beta) scaling factor. Used to re-scale the Euclidean space by 
        emphasizing different KGE components.
        
    Returns
    -------
    score: float
        Kling-Gupta efficiency.
        
    References
    ----------
    Gupta, H. V., Kling, H., Yilmaz, K. K., & Martinez, G. F. (2009). Decomposition of 
        the mean squared error and NSE performance criteria: Implications for improving 
        hydrological modelling. Journal of hydrology, 377(1-2), 80-91. 
        https://doi.org/10.1016/j.jhydrol.2009.08.003
    
    """
    # Raise if not 1-D arrays
    validate.raise_for_non_vector(y_true, y_pred)

    # Raise if not same shape
    validate.raise_for_inconsistent_shapes(y_true, y_pred)

    # Pearson correlation coefficient
    linear_correlation = np.corrcoef(y_pred, y_true)[0,1]

    # Relative variability
    relative_variability = np.std(y_pred) / np.std(y_true)

    # Relative mean
    relative_mean = np.mean(y_pred) / np.mean(y_true)

    # Scaled Euclidean distance
    euclidean_distance = np.sqrt(
        (r_scale * (linear_correlation - 1.0)) ** 2.0 + 
        (a_scale * (relative_variability - 1.0)) ** 2.0 + 
        (b_scale * (relative_mean - 1.0)) ** 2.0
        )

    # Return KGE
    return 1.0 - euclidean_distance

def compute_contingency_table(
    observed: npt.ArrayLike,
    simulated: npt.ArrayLike,
    true_positive_key: str = 'true_positive',
    false_positive_key: str = 'false_positive',
    false_negative_key: str = 'false_negative',
    true_negative_key: str = 'true_negative'
    ) -> pd.Series:
    """Compute components of a contingency table required for the evaluation of categorical 
    forecasts and simulations. Returns a pandas.Series indexed by table component. 'true_positive' 
    indicates the number of times the simulation correctly indicated True according to the 
    observations. 'false_positive' indicates the number of times the simulation incorrectly 
    indicated True according to the observations. 'false_negative' indicates the number of times 
    the simulation incorrectly indicated False according to the observations. 'true_negative' 
    indicates the number of times the simulation correctly indicated False according to the 
    observations.
        
    Parameters
    ----------
    observed: array-like, required
        Array-like of boolean values indicating observed occurrences
    simulated: array-like, required
        Array-like of boolean values indicating simulated occurrences
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
    false_negative_key: str, optional, default 'false_negative'
        Label to use for false negatives.
    true_negative_key: str, optional, default 'true_negative'
        Label to use for true negatives.
        
    Returns
    -------
    contingency_table: pandas.Series
        pandas.Series of integer values keyed to pandas.Index([true_positive_key, false_positive_key, 
            false_negative_key, true_negative_key])

    Examples
    --------
    >>> obs = [True, True, True, False, False, False, False, False, True, True]
    >>> sim = [True, True, False, False, False, False, True, False, False, False]
    >>> metrics.compute_contingency_table(obs, sim)
    true_positive     2
    false_positive    1
    false_negative    3
    true_negative     4
    dtype: int64
        
    """
    # Raise if not 1-D arrays
    validate.raise_for_non_vector(observed, simulated)

    # Raise if not same shape
    validate.raise_for_inconsistent_shapes(observed, simulated)

    # Validate boolean categorical
    observed = validate.convert_to_boolean_categorical_series(observed)
    simulated = validate.convert_to_boolean_categorical_series(simulated)

    # Cross tabulate
    ctab = pd.crosstab(observed, simulated, dropna=False)

    # Reformat
    return pd.Series({
        true_positive_key : ctab.loc[True, True],
        false_positive_key : ctab.loc[False, True],
        false_negative_key : ctab.loc[True, False],
        true_negative_key : ctab.loc[False, False]
        })

def probability_of_detection(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    true_positive_key: str = 'true_positive',
    false_negative_key: str = 'false_negative'
    ) -> float:
    """Compute probability of detection (POD), also called the "hit rate". POD 
    is the ratio of true positives to the number of observations. POD ranges 
    from 0.0 to 1.0, higher is better. Note: that this statistic is easy to 
    "hedge" if the model always indicates occurence. This statistic should 
    be considered alongside some metric of false positives, like probability 
    of false alarm or threat score.
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following 
        keys: true_positive_key, false_positive_key, false_negative_key, 
        true_negative_key; and int or float values
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_negative_key: str, optional, default 'false_negative'
        Label to use for false negatives.
        
    Returns
    -------
    POD: float
        Probability of detection.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    a = contingency_table[true_positive_key]
    c = contingency_table[false_negative_key]
    return a / (a+c)

def probability_of_false_detection(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    false_positive_key: str = 'false_positive',
    true_negative_key: str = 'true_negative'
    ) -> float:
    """Compute probability of false detection/false alarm rate (POFD/FARate). POFD 
    indicates the portion of non-occurences that were false alarms. POFD ranges from 
    0.0 to 1.0, lower is better.
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following 
        keys: true_positive_key, false_positive_key, false_negative_key, 
        true_negative_key; and int or float values
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
    true_negative_key: str, optional, default 'true_negative'
        Label to use for true negatives.
        
    Returns
    -------
    POFD: float
        Probability of false detection.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    b = contingency_table[false_positive_key]
    d = contingency_table[true_negative_key]
    return b / (b+d)

def probability_of_false_alarm(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    true_positive_key: str = 'true_positive',
    false_positive_key: str = 'false_positive'
    ) -> float:
    """Compute probability of false alarm/false alarm ratio (POFA/FARatio). POFA 
    indicates the portion of predictions or simulated values that were false alarms. 
    POFA ranges from 0.0 to 1.0, lower is better. The complement of POFA (1.0 - POFA) is 
    the 'post-agreement (PAG).'
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following 
        keys: true_positive_key, false_positive_key, false_negative_key, 
        true_negative_key; and int or float values
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
        
    Returns
    -------
    POFA: float
        Probability of false alarm.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    b = contingency_table[false_positive_key]
    a = contingency_table[true_positive_key]
    return b / (b+a)

def threat_score(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    true_positive_key: str = 'true_positive',
    false_positive_key: str = 'false_positive',
    false_negative_key: str = 'false_negative'
    ) -> float:
    """Compute threat score/critical success index (TS/CSI). CSI is the ratio 
    true positives to the sum of true positives, false positives, and false 
    negatives. CSI ranges from 0.0 to 1.0, higher is better. CSI is sensitive 
    to event frequency, in which case the equitable threat score may be more 
    suitable.
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following 
        keys: true_positive_key, false_positive_key, false_negative_key, 
        true_negative_key; and int or float values
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
    false_negative_key: str, optional, default 'false_negative'
        Label to use for false negatives.
        
    Returns
    -------
    TS: float
        Threat score.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    a = contingency_table[true_positive_key]
    b = contingency_table[false_positive_key]
    c = contingency_table[false_negative_key]
    return a / (a+b+c)

def frequency_bias(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    true_positive_key: str = 'true_positive',
    false_positive_key: str = 'false_positive',
    false_negative_key: str = 'false_negative'
    ) -> float:
    """Compute frequency bias (FBI). FBI measures the tendency of the simulation 
    or forecast to over or under-predict. FBI ranges from 0.0 to inf. A perfect 
    score is 1.0. Values less than 1.0 indicate under-prediction. Values greater 
    than 1.0 indicate over-prediction.
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following 
        keys: true_positive_key, false_positive_key, false_negative_key, 
        true_negative_key; and int or float values
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
    false_negative_key: str, optional, default 'false_negative'
        Label to use for false negatives.
        
    Returns
    -------
    FBI: float
        Frequency bias.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    a = contingency_table[true_positive_key]
    b = contingency_table[false_positive_key]
    c = contingency_table[false_negative_key]
    return (a+b) / (a+c)

def percent_correct(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    true_positive_key: str = 'true_positive',
    false_positive_key: str = 'false_positive',
    false_negative_key: str = 'false_negative',
    true_negative_key: str = 'true_negative'
    ) -> float:
    """Compute percent correct (PC). PC is the sum of both true positives and 
    true negatives compared to the total number of observations. PC is the portion 
    of correctly predicted occurences and non-occurences. PC ranges from 0.0 to 1.0, 
    higher is better.
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following 
        keys: true_positive_key, false_positive_key, false_negative_key, 
        true_negative_key; and int or float values
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
    false_negative_key: str, optional, default 'false_negative'
        Label to use for false negatives.
    true_negative_key: str, optional, default 'true_negative'
        Label to use for true negatives.
        
    Returns
    -------
    PC: float
        Percent correct.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    a = contingency_table[true_positive_key]
    b = contingency_table[false_positive_key]
    c = contingency_table[false_negative_key]
    d = contingency_table[true_negative_key]
    return (a+d) / (a+b+c+d)

def base_chance(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    true_positive_key: str = 'true_positive',
    false_positive_key: str = 'false_positive',
    false_negative_key: str = 'false_negative',
    true_negative_key: str = 'true_negative'
    ) -> float:
    """Compute base chance to hit (a_r). Base chance is the relative frequency of 
    occurences. In other words, this is the probability of scoring a "hit" or true positive 
    by chance.
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following keys: true_positive_key, false_positive_key, 
            false_negative_key, true_negative_key; and int or float values 
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
    false_negative_key: str, optional, default 'false_negative'
        Label to use for false negatives.
    true_negative_key: str, optional, default 'true_negative'
        Label to use for true negatives.
        
    Returns
    -------
    a_r: float
        Base chance to hit by chance.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    a = contingency_table[true_positive_key]
    b = contingency_table[false_positive_key]
    c = contingency_table[false_negative_key]
    d = contingency_table[true_negative_key]
    return ((a+b) * (a+c)) / (a+b+c+d)

def equitable_threat_score(
    contingency_table: Union[dict, pd.DataFrame, pd.Series],
    true_positive_key: str = 'true_positive',
    false_positive_key: str = 'false_positive',
    false_negative_key: str = 'false_negative',
    true_negative_key: str = 'true_negative'
    ) -> float:
    """Compute equitable threat score (ETS). Threat score/Critical Success Index 
    tends to yield lower scores for rare events. ETS computes a threat score, but 
    accounts for the relative frequency of scoring a true positive by chance.
        
    Parameters
    ----------
    contingency_table: dict, pandas.DataFrame, or pandas.Series, required
        Contingency table containing key-value pairs with the following 
        keys: true_positive_key, false_positive_key, false_negative_key, 
        true_negative_key; and int or float values
    true_positive_key: str, optional, default 'true_positive'
        Label to use for true positives.
    false_positive_key: str, optional, default 'false_positive'
        Label to use for false positives.
    false_negative_key: str, optional, default 'false_negative'
        Label to use for false negatives.
    true_negative_key: str, optional, default 'true_negative'
        Label to use for true negatives.
        
    Returns
    -------
    ETS: float
        Equitable threat score.
        
    """
    # Convert values to numpy scalars
    contingency_table = pd.Series(contingency_table, dtype=np.float64)

    # Compute
    a_r = base_chance(contingency_table,
        true_positive_key=true_positive_key,
        false_positive_key=false_positive_key,
        false_negative_key=false_negative_key,
        true_negative_key=true_negative_key
        )
    a = contingency_table[true_positive_key]
    b = contingency_table[false_positive_key]
    c = contingency_table[false_negative_key]
    return (a-a_r) / (a+b+c-a_r)



import pytest
from hydrotools.metrics import metrics

import pandas as pd
import numpy as np

contigency_table = {
    'true_positive': 1,
    'false_positive': 2,
    'false_negative': 3,
    'true_negative': 4
}

alt_contigency_table = {
    'TP': 1,
    'FP': 2,
    'FN': 3,
    'TN': 4
}

zero_contingency_table = {
    'true_positive': 0,
    'false_positive': 0,
    'false_negative': 0,
    'true_negative': 0
}

nan_contigency_table = {
    'true_positive': np.nan,
    'false_positive': np.nan,
    'false_negative': np.nan,
    'true_negative': np.nan
}

char_contigency_table = {
    'true_positive': '1',
    'false_positive': '2',
    'false_negative': '3',
    'true_negative': '4'
}

y_true = [1., 2., 3., 4.]
y_pred = [4., 3., 2., 1.]

z_true = [0., 0., 0., 0.]
z_pred = [0., 0., 0., 0.]

n_true = [np.nan, np.nan, np.nan, np.nan]
n_pred = [np.nan, np.nan, np.nan, np.nan]

y_true_series = pd.Series(
    data=y_true,
    index=pd.date_range(
        start="2020-01-01", 
        end="2020-01-04", 
        freq="1D"
    )
)

y_pred_series = pd.Series(
    data=y_pred,
    index=pd.date_range(
        start="2020-01-01", 
        end="2020-01-04", 
        freq="1D"
    )
)

def test_compute_contingency_table():
    obs = pd.Series([True, False, False, True, True, True,
        False, False, False, False], dtype="category")
    sim = pd.Series([True, True, True, False, False, False, 
        False, False, False, False], dtype="category")

    table = metrics.compute_contingency_table(obs, sim)

    assert table['true_positive'] == 1
    assert table['false_positive'] == 2
    assert table['false_negative'] == 3
    assert table['true_negative'] == 4

    alt_table = metrics.compute_contingency_table(obs, sim, 
        true_positive_key='TP',
        false_positive_key='FP',
        false_negative_key='FN',
        true_negative_key='TN'
        )

    assert alt_table['TP'] == 1
    assert alt_table['FP'] == 2
    assert alt_table['FN'] == 3
    assert alt_table['TN'] == 4

def test_probability_of_detection():
    POD = metrics.probability_of_detection(contigency_table)
    assert POD == (1/4)

    POD = metrics.probability_of_detection(alt_contigency_table,
        true_positive_key='TP',
        false_negative_key='FN'
        )
    assert POD == (1/4)

    with pytest.warns(RuntimeWarning):
        POD = metrics.probability_of_detection(zero_contingency_table)
        assert np.isnan(POD)

    POD = metrics.probability_of_detection(nan_contigency_table)
    assert np.isnan(POD)

    POD = metrics.probability_of_detection(char_contigency_table)
    assert POD == (1/4)

def test_probability_of_false_detection():
    POFD = metrics.probability_of_false_detection(contigency_table)
    assert POFD == (2/6)

    POFD = metrics.probability_of_false_detection(char_contigency_table)
    assert POFD == (2/6)

    POFD = metrics.probability_of_false_detection(alt_contigency_table,
        false_positive_key='FP',
        true_negative_key='TN'
        )
    assert POFD == (2/6)

    with pytest.warns(RuntimeWarning):
        POFD = metrics.probability_of_false_detection(zero_contingency_table)
        assert np.isnan(POFD)

    POFD = metrics.probability_of_false_detection(nan_contigency_table)
    assert np.isnan(POFD)

def test_probability_of_false_alarm():
    POFA = metrics.probability_of_false_alarm(contigency_table)
    assert POFA == (2/3)

    POFA = metrics.probability_of_false_alarm(char_contigency_table)
    assert POFA == (2/3)

    POFA = metrics.probability_of_false_alarm(alt_contigency_table,
        true_positive_key='TP',
        false_positive_key='FP'
        )
    assert POFA == (2/3)

    with pytest.warns(RuntimeWarning):
        POFA = metrics.probability_of_false_alarm(zero_contingency_table)
        assert np.isnan(POFA)

    POFA = metrics.probability_of_false_alarm(nan_contigency_table)
    assert np.isnan(POFA)

def test_threat_score():
    TS = metrics.threat_score(contigency_table)
    assert TS == (1/6)

    TS = metrics.threat_score(char_contigency_table)
    assert TS == (1/6)

    TS = metrics.threat_score(alt_contigency_table,
        true_positive_key='TP',
        false_positive_key='FP',
        false_negative_key='FN'
        )
    assert TS == (1/6)

    with pytest.warns(RuntimeWarning):
        TS = metrics.threat_score(zero_contingency_table)
        assert np.isnan(TS)

    TS = metrics.threat_score(nan_contigency_table)
    assert np.isnan(TS)

def test_frequency_bias():
    FBI = metrics.frequency_bias(contigency_table)
    assert FBI == (3/4)

    FBI = metrics.frequency_bias(char_contigency_table)
    assert FBI == (3/4)

    FBI = metrics.frequency_bias(alt_contigency_table,
        true_positive_key='TP',
        false_positive_key='FP',
        false_negative_key='FN'
        )
    assert FBI == (3/4)

    with pytest.warns(RuntimeWarning):
        FBI = metrics.frequency_bias(zero_contingency_table)
        assert np.isnan(FBI)

    FBI = metrics.frequency_bias(nan_contigency_table)
    assert np.isnan(FBI)

def test_percent_correct():
    PC = metrics.percent_correct(contigency_table)
    assert PC == (5/10)

    PC = metrics.percent_correct(char_contigency_table)
    assert PC == (5/10)

    PC = metrics.percent_correct(alt_contigency_table,
        true_positive_key='TP',
        false_positive_key='FP',
        false_negative_key='FN',
        true_negative_key='TN'
        )
    assert PC == (5/10)

    with pytest.warns(RuntimeWarning):
        PC = metrics.percent_correct(zero_contingency_table)
        assert np.isnan(PC)

    PC = metrics.percent_correct(nan_contigency_table)
    assert np.isnan(PC)

def test_base_chance():
    a_r = metrics.base_chance(contigency_table)
    assert a_r == (12/10)

    a_r = metrics.base_chance(char_contigency_table)
    assert a_r == (12/10)

    a_r = metrics.base_chance(alt_contigency_table,
        true_positive_key='TP',
        false_positive_key='FP',
        false_negative_key='FN',
        true_negative_key='TN'
        )
    assert a_r == (12/10)

    with pytest.warns(RuntimeWarning):
        a_r = metrics.base_chance(zero_contingency_table)
        assert np.isnan(a_r)

    a_r = metrics.base_chance(nan_contigency_table)
    assert np.isnan(a_r)

def test_equitable_threat_score():
    ETS = metrics.equitable_threat_score(contigency_table)
    assert np.isclose(ETS, (-0.2/4.8), atol=0.000001)

    ETS = metrics.equitable_threat_score(char_contigency_table)
    assert np.isclose(ETS, (-0.2/4.8), atol=0.000001)

    ETS = metrics.equitable_threat_score(alt_contigency_table,
        true_positive_key='TP',
        false_positive_key='FP',
        false_negative_key='FN',
        true_negative_key='TN'
        )
    assert np.isclose(ETS, (-0.2/4.8), atol=0.000001)

    with pytest.warns(RuntimeWarning):
        ETS = metrics.equitable_threat_score(zero_contingency_table)
        assert np.isnan(ETS)

    ETS = metrics.equitable_threat_score(nan_contigency_table)
    assert np.isnan(ETS)

def test_mean_absolute_error():
    MAE = metrics.mean_error(y_true, y_pred)
    assert MAE == 2.0

def test_mean_error():
    MSE = metrics.mean_error(y_true, y_pred, power=2.0)
    assert MSE == 5.0

    RMSE = metrics.mean_error(y_true, y_pred, power=2.0, root=True)
    assert RMSE == np.sqrt(5.0)

def test_nash_sutcliffe_efficiency():
    NSE = metrics.nash_sutcliffe_efficiency(y_true, y_pred)
    assert NSE == -3.0
    
    NNSE = metrics.nash_sutcliffe_efficiency(y_true, y_pred, 
        normalized=True)
    assert NNSE == 0.2
    
    NSEL = metrics.nash_sutcliffe_efficiency(np.exp(y_true), 
        np.exp(y_pred), log=True)
    assert NSEL == -3.0
    
    NNSEL = metrics.nash_sutcliffe_efficiency(np.exp(y_true), 
        np.exp(y_pred), log=True, normalized=True)
    assert NNSEL == 0.2

def test_zero_mean_error():
    MSE = metrics.mean_error(z_true, z_pred, power=2.0)
    assert MSE == 0.0

    RMSE = metrics.mean_error(z_true, z_pred, power=2.0, root=True)
    assert RMSE == 0.0

def test_nan_mean_error():
    MSE = metrics.mean_error(n_true, n_pred, power=2.0)
    assert np.isnan(MSE)

    RMSE = metrics.mean_error(n_true, n_pred, power=2.0, root=True)
    assert np.isnan(RMSE)

def test_zero_nash_sutcliffe_efficiency():
    with pytest.warns(RuntimeWarning):
        NSE = metrics.nash_sutcliffe_efficiency(z_true, z_pred)
        assert np.isnan(NSE)
        
        NNSE = metrics.nash_sutcliffe_efficiency(z_true, z_pred, 
            normalized=True)
        assert np.isnan(NNSE)
       
        NSEL = metrics.nash_sutcliffe_efficiency(np.exp(z_true), 
            np.exp(z_pred), log=True)
        assert np.isnan(NSEL)
        
        NNSEL = metrics.nash_sutcliffe_efficiency(np.exp(z_true), 
            np.exp(z_pred), log=True, normalized=True)
        assert np.isnan(NNSEL)

def test_nan_nash_sutcliffe_efficiency():
    NSE = metrics.nash_sutcliffe_efficiency(n_true, n_pred)
    assert np.isnan(NSE)
    
    NNSE = metrics.nash_sutcliffe_efficiency(n_true, n_pred, 
        normalized=True)
    assert np.isnan(NNSE)
    
    NSEL = metrics.nash_sutcliffe_efficiency(np.exp(n_true), 
        np.exp(n_pred), log=True)
    assert np.isnan(NSEL)
    
    NNSEL = metrics.nash_sutcliffe_efficiency(np.exp(n_true), 
        np.exp(n_pred), log=True, normalized=True)
    assert np.isnan(NNSEL)

def test_kling_gupta_efficiency():
    # Default (inverse case)
    KGE = metrics.kling_gupta_efficiency(y_true, y_pred)
    expected = -1.0
    assert np.isclose(KGE, expected)

    # Half-scale
    KGE = metrics.kling_gupta_efficiency(y_true, np.array(y_true) * 0.5)
    expected = (1.0 - np.sqrt(0.5))
    assert np.isclose(KGE, expected)

    # Double-scale
    KGE = metrics.kling_gupta_efficiency(y_true, np.array(y_true) * 2.0)
    expected = (1.0 - np.sqrt(2.0))
    assert np.isclose(KGE, expected)

    # Identity
    KGE = metrics.kling_gupta_efficiency(y_true, np.array(y_true) * 1.0)
    expected = 1.0
    assert np.isclose(KGE, expected)

    # Scaling parameters
    KGE = metrics.kling_gupta_efficiency(y_true, np.array(y_true) * 0.25,
        r_scale=0.5, a_scale=0.25, b_scale=0.25)
    expected = (1.0 - np.sqrt(9.0/128.0))
    assert np.isclose(KGE, expected)

def test_coefficient_of_persistence():
    # Default
    COP = metrics.coefficient_of_persistence(y_true, y_pred)
    expected = (1.0 - 11.0/3.0)
    assert np.isclose(COP, expected)

    # Test with series
    COP = metrics.coefficient_of_persistence(y_true_series, y_pred_series)
    expected = (1.0 - 11.0/3.0)
    assert np.isclose(COP, expected)

    # Identity
    COP = metrics.coefficient_of_persistence(y_true, np.array(y_true) * 1.0)
    expected = 1.0
    assert np.isclose(COP, expected)

    # Lag
    COP = metrics.coefficient_of_persistence(y_true, y_pred, lag=2)
    expected = -0.25
    assert np.isclose(COP, expected)

    # Power
    COP = metrics.coefficient_of_persistence(y_true, y_pred, power=3)
    expected = (1.0 - 29.0 / 3.0)
    assert np.isclose(COP, expected)

    # Normalized
    COP = metrics.coefficient_of_persistence(y_true, y_pred, normalized=True)
    expected = 1.0 / (2.0 - (1.0 - 11.0/3.0))
    assert np.isclose(COP, expected)

    # Log
    COP = metrics.coefficient_of_persistence(y_true, y_pred, log=True)
    expected = -2.09313723301667
    assert np.isclose(COP, expected)

def test_coefficient_of_extrapolation():
    # Default
    v = [1, 2, 4, 5]
    COE = metrics.coefficient_of_extrapolation(v, y_pred)
    expected = -3.0
    assert np.isclose(COE, expected)

    # Test with series
    s = pd.Series(data=v, index=y_true_series.index)
    COE = metrics.coefficient_of_extrapolation(s, y_pred_series)
    expected = -3.0
    assert np.isclose(COE, expected)

    # Identity
    COE = metrics.coefficient_of_extrapolation(y_true, np.array(y_true) * 1.0)
    expected = 1.0
    assert np.isclose(COE, expected)

    # Power
    COE = metrics.coefficient_of_extrapolation(v, y_pred, power=1.5)
    expected = -1.82842712474619
    assert np.isclose(COE, expected)

    # Normalized
    COE = metrics.coefficient_of_extrapolation(v, y_pred, normalized=True)
    expected = 0.2
    assert np.isclose(COE, expected)

    # Log
    COE = metrics.coefficient_of_extrapolation(v, y_pred, log=True)
    expected = -2.19567503891363
    assert np.isclose(COE, expected)

def test_mean_squared_error():
    MSE = metrics.mean_squared_error(y_true, y_pred)
    assert MSE == 5.0

def test_root_mean_squared_error():
    RMSE = metrics.root_mean_squared_error(y_true, y_pred)
    assert RMSE == np.sqrt(5.0)

def test_volumetric_efficiency():
    VE = metrics.volumetric_efficiency(y_true, y_pred)
    assert np.isclose(VE, 0.2)
